#	Masrv
#	Copyright (C) 2007 Michal Vaner (vorner@ucw.cz)
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	he Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import glob
from commandable import commandable
import re
import copy

class bind:
	def __init__( self, x, y, z, object ):
		self.object = object
		self.z = z
		self.pos = ( x, y )
		self.__speeds = [ ( x, y ) ]
		self.__hooks = []

	def position( self, x, y, derivate ):
		if derivate == 0: self.pos = ( x, y )
		while( len( self.__speeds ) <= derivate ): self.__speeds.append( ( 0, 0 ) )
		self.__speeds[ derivate ] = ( x, y )
		while( self.__speeds[ -1 ] == ( 0, 0 ) ): self.__speeds.pop()
		if not self.__speeds:
			self.speeds.append( ( 0, 0 ) )

	def update( self, timestep ):
		self.__speeds.reverse()
		last = ( 0, 0 )
		updated = []
		for ( x, y ) in self.__speeds:
			( lastx, lasty ) = last
			n = ( x + ( lastx * timestep / 1000.0 ), y + ( lasty * timestep / 1000.0 ) )
			updated.insert( 0, n )
			last = n
		self.__speeds = updated
		( self.x, self.y ) = self.__speeds[ 0 ]
		for h in self.__hooks:
			( x, y, hix, hiy, id ) = h
			def comp( z, az, hiz ):
				if hiz == 1:
					return z <= az
				elif hiz == -1:
					return z >= az
				else:
					return True
			( mx, my ) = map( comp, ( x, y ), ( self.x, self.y ), ( hix, hiy ) )
			if mx and my:
				self.__hooks.remove( h )
				glob.hooks.trigger( id )

	def posHook( self, x, y, hix, hiy, id ):
		self.__hooks.append( ( x, y, hix, hiy, id ) )

	def getState( self, flipx, flipy, moves ):
		( multx, multy ) = ( 1, 1 )
		if flipx: multx = -1
		if flipy: multy = -1
		def sum( l, r ):
			if l:
				( lx, ly ) = l
			else:
				( lx, ly ) = ( 0, 0 )
			if r:
				( rx, ry ) = r
			else:
				( rx, ry ) = ( 0, 0 )
			return ( rx + multx * lx, ry + multy * ly )
		nmoves = map( sum, self.__speeds, moves )
		result = map( lambda x: ( x[ 0 ], x[ 1 ] + str( self ) ) + x[ 2: ], self.object.getState( flipx, flipy, nmoves ) )
		return result

class object:
	def __init__( self ):
		self.flipX = False
		self.flipY = False
		self.__subs = {}
		self.__nonames = []

	def insert( self, name, object, x, y, z ):
		if name in self.__subs:
			self.__nonames.append( self.__subs[ name ] )
		self.__subs[ name ] = bind( x, y, z, object )

	def remove( self, name ):
		del self.__subs[ name ]

	def position( self, name, x, y, derivate ):
		self.__subs[ name ].position( x, y, derivate )

	def setZ( self, name, z ):
		self.__subs[ name ].z = z

	def update( self, timestep ):
		for o in self.__subs.values() + self.__nonames:
			o.update( timestep )
			o.object.update( timestep )

	def posHook( self, name, x, y, hix, hiy, id ):
		self.__subs[ name ].posHook( x, y, hix, hiy, id )

	def getState( self, flipx, flipy, moves ):
		result = []
		if self.flipX:#Turn it
			flipx = not flipx
		if self.flipY:#Same with y
			flipy = not flipy
		list = self.__subs.values() + self.__nonames
		list.sort( cmp, lambda x: x.z, True )
		for b in list:
			result.extend( b.getState( flipx, flipy, moves ) )#Descend inside with the right moves and so
		result.extend( self.localState( flipx, flipy, moves ) )
		return result

	def localState( self, flipx, flipy, moves ):
		return []

class imgObject( object ):
	def __init__( self, image, images ):
		object.__init__( self )
		self.__images = [ image ]
		self.__time = 100000 #Something large enough, it has no effect with 1 image anyway
		self.__size = images.size( image )
	def addImage( self, image, time ):
		if time != None:
			self.__time = time
		self.__images.append( image )
	def localState( self, flipx, flipy, moves ):
		mmoves = copy.copy( moves )
		if flipx:
			( x, y ) = mmoves[ 0 ]
			mmoves[ 0 ] = ( x - self.__size[ 0 ], y )
		if flipy:
			( x, y ) = mmoves[ 0 ]
			mmoves[ 0 ] = ( x, y - self.__size[ 1 ] )
		return [ ( "i", "", self.__images, self.__time, self.__size[ 0 ], self.__size[ 1 ], flipx, flipy, mmoves ) ]

class lineObject( object ):
	def __init__( self, xdelta, ydelta, width, r, g, b ):
		object.__init__( self )
		self.__xdelta = xdelta
		self.__ydelta = ydelta
		self.__width = width
		self.__color = ( r, g, b )
	def localState( self, flipx, flipy, moves ):
		( xd, yd ) = ( self.__xdelta, self.__ydelta )
		if flipx: xd *= -1
		if flipy: yd *= -1
		return [ ( "l", "", self.__color, xd, yd, self.__width, moves ) ]

class modeler( commandable ):
	"""The modeler - it takes care of having all things on the right place, moving them and so on"""
	def __init__( self, source, images ):
		commandable.__init__( self, { "def":self.__defObj, "undef":self.__undefObj, "placeImg":self.__placeImg, "addImg":self.__addImg, "placeLn":self.__placeLn, "placeObj":self.__placeObj, "unplace":self.__unplaceObj, "position":self.__positionObj, "setZ":self.__setZ, "flipX":self.__flipX, "flipY":self.__flipY } )
		glob.info( "Creating the modeler" )
		self.__source = source
		rootObj = object()
		self.__objects = { "root":rootObj }
		self.__root = rootObj
		self.changed = False
		self.__lasttimestamp = glob.timestamp()
		self.__images = images

	def __defObj( self, args, id ):
		obj = object()
		self.__objects[ args[ 0 ] ] = obj

	def __undefObj( self, args, id ):
		del self.__objects[ args[ 0 ] ]

	def __placeImg( self, args, id ):
		( where, img, name ) = args[ :3 ]
		( x, y ) = map( float, args[ 3:5 ] )
		z = int( args[ 5 ] )
		obj = imgObject( img, self.__images )
		self.__objects[ where ].insert( name, obj, x, y, z )
		if len( args ) >= 7:
			self.__objects[ args[ 6 ] ] = obj
		self.changed = True

	def __addImg( self, args, id ):
		( name, img ) = args[ :2 ]
		if len( args ) >= 3:
			time = int( args[ 2 ] )
		else:
			time = None
		self.__objects[ name ].addImage( img, time )
		self.changed = True

	def __placeLn( self, args, id ):
		( where, color, width, name ) = args[ :4 ]
		width = int( width )
		( x1, y1, x2, y2 ) = map( float, args[ 4:8 ] )
		z = int( args[ 8 ] )
		( r, g, b ) = map( lambda x: int( x, 16 ), re.findall( "..", color ) )
		obj = lineObject( int( x2 - x1 ), int( y2 - y1 ), width, r, g, b )
		self.__objects[ where ].insert( name, obj, x1, y1, z )
		if len( args ) >= 10:
			self.__objects[ args[ 9 ] ] = obj
		self.changed = True

	def __placeObj( self, args, id ):
		( where, what, name ) = args[ :3 ]
		( x, y ) = map( float, args[ 3:5 ] )
		z = int( args[ 5 ] )
		obj = self.__objects[ what ]
		self.__objects[ where ].insert( name, obj, x, y, z )
		self.changed = True

	def __unplaceObj( self, args, id ):
		( where, what ) = args[ :2 ]
		self.__objects[ where ].remove( what )
		self.changed = True

	def __positionObj( self, args, id ):
		( where, what ) = args[ :2 ]
		coords = map( float, args[ 2: ] )
		coords.append( 0 )
		( x, y, d ) = coords[ :3 ]
		d = int( d )
		self.__objects[ where ].position( what, x, y, d )
		self.changed = True

	def __setZ( self, args, id ):
		( where, what ) = args[ :2 ]
		z = int( args[ 2 ] )
		self.__objects[ where ].setZ( what, z )
		self.changed = True

	def __flipX( self, args, id ):
		obj = self.__objects[ args[ 0 ] ]
		obj.flipX = not obj.flipX
		self.changed = True

	def __flipY( self, args, id ):
		obj = self.__objects[ args[ 0 ] ]
		obj.flipY = not obj.flipY
		self.changed = True

	def posHook( self, where, what, x, y, hix, hiy, id ):
		self.__objects[ where ].posHook( what, x, y, hix, hiy, id )

	def update( self, timestamp ):
		timestep = timestamp - self.__lasttimestamp
		self.__lasttimestamp = timestamp
		self.__root.update( timestep )
		return self.changed
	def getState( self ):
		return self.__root.getState( False, False, [ ( 0, 0 ) ] )
